/*
 * Part of "Intel(R) AMT Management Interface" Linux driver 
 *
 * Copyright (c) 2003 - 2007 Intel Corp.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */

#ifndef _HECI_DATA_STRUCTURES_H_
#define _HECI_DATA_STRUCTURES_H_

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/pci.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/module.h>
#include <linux/aio.h>
#include <linux/types.h>

/**
 * error code definition 
 */
#define     ESUCCESS                     0
#define     ESLOTS_OVERFLOW              1
#define     ECORRUPTED_MESSAGE_HEADER    1000
#define     ECOMPLETE_MESSAGE            1001

/**
 * Number of queue lists used by this driver
 */
#define NUMBER_OF_LISTS        7

#define LEGACY_MTU 4160
#pragma pack(1)


/**
 * HECI HW Section
 */

/* HECI addresses and defines */
#define H_CB_WW    0
#define H_CSR      4
#define ME_CB_RW   8
#define ME_CSR_HA  0xC


/* register bits - H_CSR */

#define H_CBD             0xFF000000
#define H_CBWP            0x00FF0000
#define H_CBRP            0x0000FF00
#define H_RST             0x00000010
#define H_RDY             0x00000008
#define H_IG              0x00000004
#define H_IS              0x00000002
#define H_IE              0x00000001


/* register bits - ME_CSR_HA */
#define ME_CBD_HRA        0xFF000000
#define ME_CBWP_HRA       0x00FF0000
#define ME_CBRP_HRA       0x0000FF00
#define ME_RST_HRA        0x00000010
#define ME_RDY_HRA        0x00000008
#define ME_IG_HRA         0x00000004
#define ME_IS_HRA         0x00000002
#define ME_IE_HRA         0x00000001

/**
 *  heci driver use additional char device for legacy mode 
 */
#define  MINORS_COUNT           		2

#define  LEGACY_MINOR_NUMBER			0
#define  HECI_MINOR_NUMBER			    1
#define  MAX_OPEN_HANDLE_COUNT          253
/**
 * debug kernel print macro define
 */
#define	INFO(format, arg...)	        printk(KERN_INFO     "%s: " format,THIS_MODULE->name, ## arg)
#define	ERR(format, arg...)	            printk(KERN_ERR      "%s: " format,THIS_MODULE->name, ## arg)
#define WARN(format, arg...)            printk(KERN_WARNING  "%s: " format,THIS_MODULE->name, ## arg)


/* Module Parameters */
#define	DEF_PARM(type, name, init, perm, desc)	\
	type name = (init);	\
	MODULE_PARM_DESC(name, desc);	\
	module_param(name, type, perm)

extern int debug;

#define DBG(format, arg...) do {if (debug) \
printk(KERN_ERR "%s: " format , __FUNCTION__ , ## arg); \
} while (0)

#ifdef HECI_DEBUG
#define assert(expr) do {} while (0)
#else
#define assert(expr) \
    if(!(expr)) {                                   \
    printk( "Assertion failed! %s,%s,%s,line=%d\n", \
    #expr,__FILE__,__FUNCTION__,__LINE__);          \
    }
#endif

/** 
 * time to wait event 
 */
#define HECI_INTEROP_TIMEOUT    (HZ * 7)

/**
 * watch dog definition  
 */
#define HECI_WATCHDOG_DATA_SIZE         16
#define HECI_START_WD_DATA_SIZE         20
#define HECI_WD_PARAMS_SIZE             4


#define HECI_WD_HOST_CLIENT_ID          1
#define HECI_LEGACY_HOST_CLIENT_ID      2
typedef enum {
#undef FALSE
	FALSE = 0,
#undef TRUE
	TRUE = 1
} boolean_t;


typedef struct guid {
	uint32_t data1;
	uint16_t data2;
	uint16_t data3;
	uint8_t data4[8];
} guid_t;

/* File state */
enum file_state {
	HECI_FILE_INITIALIZING,
	HECI_FILE_CONNECTING,
	HECI_FILE_CONNECTED,
	HECI_FILE_DISCONNECTING,
	HECI_FILE_DISCONNECTED
};

/* HECI states */
typedef enum {
	HECI_INITIALIZING = 0,
	HECI_ENABLED,
	HECI_RESETING,
	HECI_DISABLED,
	HECI_RECOVERING_FROM_RESET,
	HECI_POWER_DOWN,
	HECI_POWER_UP
} heci_state_t;

typedef enum {
	HECI_LEGACY_IDLE,
	HECI_LEGACY_WRITING,
	HECI_LEGACY_FLOW_CONTROL,
	HECI_LEGACY_READING,
	HECI_LEGACY_READ_COMPLETE
} legacy_state_t;

typedef enum {
	HECI_IDLE,
	HECI_WRITING,
	HECI_WRITE_COMPLETE,
	HECI_FLOW_CONTROL,
	HECI_READING,
	HECI_READ_COMPLETE
} heci_file_transaction_state_t;

/* HECI CB */
typedef enum {
	HECI_READ = 0,
	HECI_WRITE,
	HECI_IOCTL,
	HECI_OPEN,
	HECI_CLOSE
} heci_cb_major_t;

/* HECI user data struct */
struct heci_message_data {
	uint32_t size;
	char *data;
};
#define SECOND_TO_MILLI                 1000
#define SECOND_TO_MICRO                 SECOND_TO_MILLI * 1000
#define SECOND_TO_100NANO               SECOND_TO_MICRO * 10

#define CONNECT_TIMEOUT                 3	/* at least 2 seconds */

#define LEGACY_STALL_TIMER              12	/* seconds */

struct heci_cb_private {
	struct list_head cb_list;
	heci_cb_major_t major_file_operations;
	void *file_private;
	struct heci_message_data request_buffer;
	struct heci_message_data response_buffer;
	unsigned long information;
	struct file *file_object;
};

/* Private file struct */
typedef struct heci_file_private {
	struct list_head link;
	struct file *file;
	enum file_state state;
	wait_queue_head_t tx_wait;
	wait_queue_head_t rx_wait;
	wait_queue_head_t wait;
	spinlock_t file_lock;
	spinlock_t read_io_lock;
	spinlock_t write_io_lock;
	boolean_t read_pending;
	int status;
	/* ID of client connected */
	uint8_t host_client_id;
	uint8_t me_client_id;
	uint8_t flow_control_credentials;
	uint8_t timer_count;
	heci_file_transaction_state_t reading_state;
	heci_file_transaction_state_t writing_state;
	struct heci_cb_private *read_cb;
} heci_file_private_t;

struct io_heci_list {
	struct heci_cb_private heci_cb;
	int status;
	struct iamt_heci_device *device_extension;
};

typedef struct heci_driver_version {
	uint8_t major;
	uint8_t minor;
	uint8_t hotfix;
	uint16_t build;
} heci_driver_version_t;

typedef struct heci_client {
	uint32_t max_message_length;
	uint8_t protocol_version;
} heci_client_t;
/*
 *  HECI BUS Interface Section
 */
typedef struct heci_message_header {
	uint32_t me_address:8;
	uint32_t host_address:8;
	uint32_t length:9;
	uint32_t reserved:6;
	uint32_t message_complete:1;
} heci_message_header_t;


typedef struct hbm_command {
	uint8_t command:7;
	uint8_t is_response:1;
} hbm_command_t;


typedef struct heci_bus_message {
	hbm_command_t command;
	uint8_t command_specific_data[];
} heci_bus_message_t;

typedef struct hbm_version {
	uint8_t minor_version;
	uint8_t major_version;
} hbm_version_t;

typedef struct hbm_host_version_request {
	hbm_command_t command;
	uint8_t reserved;
	hbm_version_t host_version;
} hbm_host_version_request_t;

typedef struct hbm_host_version_response {
	hbm_command_t command;
	boolean_t host_version_supported;
	hbm_version_t me_max_version;
} hbm_host_version_response_t;

typedef struct hbm_host_stop_request {
	hbm_command_t command;
	uint8_t reason;
	uint8_t reserved[2];
} hbm_host_stop_request_t;

typedef struct hbm_host_stop_response {
	hbm_command_t command;
	uint8_t reserved[3];
} hbm_host_stop_response_t;

typedef struct hbm_me_stop_request {
	hbm_command_t command;
	uint8_t reason;
	uint8_t reserved[2];
} hbm_me_stop_request_t;

typedef struct hbm_host_enumeration_request {
	hbm_command_t command;
	uint8_t reserved[3];
} hbm_host_enumeration_request_t;

typedef struct hbm_host_enumeration_response {
	hbm_command_t command;
	uint8_t reserved[3];
	uint8_t valid_addresses[32];
} hbm_host_enumeration_response_t;

typedef struct heci_client_properties {
	guid_t protocol_name;
	uint8_t protocol_version;
	uint8_t max_number_of_connections;
	uint8_t fixed_address;
	uint8_t single_receive_buffer;
	uint32_t max_message_length;
} heci_client_properties_t;

typedef struct hbm_host_client_properties_request {
	hbm_command_t command;
	uint8_t address;
	uint8_t reserved[2];
} hbm_host_client_properties_request_t;


typedef struct hbm_host_client_properties_response {
	hbm_command_t command;
	uint8_t address;
	uint8_t status;
	uint8_t reserved[1];
	heci_client_properties_t client_properties;
} hbm_host_client_properties_response_t;

typedef struct hbm_client_connect_request {
	hbm_command_t command;
	uint8_t me_address;
	uint8_t host_address;
	uint8_t reserved;
} hbm_client_connect_request_t;

typedef struct hbm_client_connect_response {
	hbm_command_t command;
	uint8_t me_address;
	uint8_t host_address;
	uint8_t status;
} hbm_client_connect_response_t;

typedef struct hbm_client_disconnect_request {
	hbm_command_t command;
	uint8_t me_address;
	uint8_t host_address;
	uint8_t reserved[1];
} hbm_client_disconnect_request_t;

typedef struct hbm_flow_control {
	hbm_command_t command;
	uint8_t me_address;
	uint8_t host_address;
	uint8_t reserved[5];
} hbm_flow_control_t;

typedef struct heci_me_client {
	heci_client_properties_t properteis;
	uint8_t client_id;
	uint8_t flow_control_credentials;
} heci_me_client_t;

/* private device struct */
typedef struct iamt_heci_device {
	struct pci_dev *pdev;	/* pointer to pci device struct */
	/*
	 * lists of queues   
	 */
	struct io_heci_list *io_list_array[NUMBER_OF_LISTS];	/* array of pointers to  aio lists */
	struct io_heci_list read_list;	/* driver  read queue */
	struct io_heci_list write_list;	/* driver write queue */
	struct io_heci_list write_waiting_list;	/* driver write waiting queue */
	struct io_heci_list control_write_list;	/* driver managed write IOCTL list */
	struct io_heci_list control_read_list;	/* driver managed read IOCTL list */
	struct io_heci_list pthi_cmd_list;	/* driver managed PTHI list for cmd waiting */
	struct io_heci_list pthi_read_complete_list;	/* driver managed PTHI list for  read completed pthi command data */
	/*
	 * list of files    
	 */
	struct list_head file_list;
	/*
	 * memory of device   
	 */
	unsigned int mem_base;
	unsigned int mem_length;
	char *mem_addr;
	/* 
	 * lock for the device 
	 */
	spinlock_t device_lock;
	spinlock_t extra_lock;
	/* 
	 * intterupts  
	 */
	int irq;
	struct work_struct work;
	boolean_t received_message;

	struct timer_list timer;
	struct timer_list wd_timer;
	/*
	 * hw states of host and fw(ME)
	 */
	uint32_t host_hw_state;
	uint32_t me_hw_state;
	/*
	 * waiting queue for receive message from FW  
	 */
	wait_queue_head_t wait_received_message;
	wait_queue_head_t wait_stop_wd;
	/*
	 * heci device  states 
	 */
	heci_state_t heci_state;
	boolean_t stop;
     /**
      * virtual void GetParam(const char* UserParam);
      * read write messages to/from heci fw 
      */
	uint32_t extra_write_index;
	uint32_t read_message_buffer[128];	/* used for control messages */
	uint32_t write_message_buffer[128];	/* used for control messages */
	uint32_t extra_message_buffer[8];	/* for control responses    */
	uint32_t read_message_header;

	hbm_version_t version;

	boolean_t host_buffer_is_empty;
	heci_file_private_t wd_file_extension;
	heci_me_client_t *me_clients;	/* Note: memory has to be allocated */
	uint8_t heci_me_clients[32];	/* list of existing clients */
	uint8_t num_heci_me_clients;
	uint8_t heci_host_clients[32];	/* list of existing clients */
	uint8_t current_host_client_id;

	boolean_t wd_pending;
	boolean_t wd_stoped;
	uint16_t wd_timeout;	/* seconds ((wd_data[1] << 8) + wd_data[0]) */
	unsigned char wd_data[HECI_START_WD_DATA_SIZE];
	unsigned char *asf_wd_data;
	unsigned long *wd_timer_tic;

	uint16_t wd_due_counter;
	boolean_t asf_mode;

	/* maybe this is not required */
	struct file *legacy_file_object;
	heci_file_private_t legacy_file_extension;
	boolean_t legacy_ioctl;
	boolean_t legacy_canceled;
	uint32_t legacy_timer;
	uint32_t legacy_stall_timer;
	unsigned char legacy_message_buffer[LEGACY_MTU];
	uint32_t legacy_message_buffer_size;
	uint32_t legacy_message_buffer_index;
	boolean_t legacy_flow_control_pending;
	legacy_state_t legacy_state;

	struct heci_cb_private *legacy_current_cb;
	uint8_t write_hang;
	boolean_t need_reset;
	long open_handle_count;

} iamt_heci_device_t;

/**
 * read_heci_register - Read a byte from the heci device
 * @device: the device structure
 * @offset: offset from which to read the data
 * 
 * Return:
 * the byte read.
 */
uint32_t read_heci_register(iamt_heci_device_t * device,
			    unsigned long offset);

/**
 * write_heci_register - Write  4 bytes to the heci device
 * @device: the device structure
 * @offset: offset from which to write the data
 * 
 * @value: the byte to write
 */
void write_heci_register(iamt_heci_device_t * device, unsigned long offset,
			 uint32_t value);

#endif				/* _HECI_DATA_STRUCTURES_H_ */
